# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import mxnet as mx
from mxnet import nd, gluon
import numpy as np

from dpareto.models.adult.lr.base import AdultLrBase


class DpAdultLrAdam(AdultLrBase):
    def __init__(self, hyperparams, options={}):
        options['optimizer_name'] = 'dp_adam'
        super(DpAdultLrAdam, self).__init__(hyperparams, options)

        self._input_layer = 123
        self._hidden_layers = []
        self._output_layer = 1

        self._name = 'dp_adult_lr_adam'


# Small run for testing purposes.
def main():
    print("Running dp_adult_lr_adam's main().")

    # Fixing some combo of these random seeds is useful for debugging. I'll fix them all to be safe.
    import random
    random_seed = 112358
    random.seed(random_seed)
    np.random.seed(random_seed)
    mx.random.seed(random_seed)

    # Some default hyperparameter setting.
    hyperparams = {'epochs': 1,
                   'lot_size': 64,
                   'lr': 0.05,
                   'l2_clipping_bound': 2.0,
                   'z': 1.0,
                   'fixed_delta': 1e-5,
                   'beta_1': 0.9,
                   'beta_2': 0.999}
    print("hyperparams: {}".format(hyperparams))

    options = {'use_gpu': False, 'verbose': True, 'accumulate_privacy': True, 'debugging': True}

    instance = DpAdultLrAdam(hyperparams, options)
    priv, acc = instance.run()
    print("Instance privacy: {}".format((priv, hyperparams['fixed_delta'])))
    print("Instance accuracy: {}".format(acc))


if __name__ == "__main__":
    main()



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import gc
import matplotlib.pyplot as plt
import multiprocessing as mp
from multiprocessing import Pool
import os
import pickle
import psutil
import time
import argparse

import gpflow
import gpflowopt
import numpy as np
import scipy as sp
import tensorflow as tf

from dpareto.utils.object_io import save_object
from dpareto.utils.random_seed_setter import set_random_seed

os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'  # disables TensorFlow debugging info dump that appears on Windows machines
#os.environ['CUDA_VISIBLE_DEVICES'] = '-1'  # forces TensorFlow to use CPU instead of defaulting to GPU

class HypervolumeImprovementHarness:
    def __init__(self, model_class, problem_name, initial_data_options, anti_ideal_point, optimization_domain, fixed_hyperparams, instance_options={}, num_instances=0, num_replications=1, num_workers=1, plot_options={}, output_base_dir=''):
        mp.set_start_method('spawn', force=True)  # workaround for weird matplotlib+multiprocessing compatibility issue
        
        self._model_class = model_class
        self._problem_name = problem_name
        self._num_instances = num_instances
        self._num_replications = num_replications

        # This is used as an ordering for consistently converting the hyperparams dict of
        # hyperparam-name-string -> hyperparam-value that we use into a numpy array of hyperparam
        # values that GPFlowOpt uses.
        self._optimizable_hyperparams_names = list(optimization_domain.keys())

        parser = argparse.ArgumentParser()
        parser.add_argument('--init-data-file')
        args = parser.parse_args()

        # Loads initial data into numpy arrays from file or generates it.
        if args.init_data_file is not None:
            initial_data_options["filename"] = args.init_data_file
        initial_data = self._get_initial_points(initial_data_options)
        self._initial_hyperparams_vals = initial_data[0]
        self._initial_privacy_vals = initial_data[1]
        self._initial_utility_vals = initial_data[2]
        self._max_initial_utility_variance = initial_data[3]

        # Transforms the initial data into a format that GPFlowOpt can use.
        transformed_data = self._create_initial_transformations(initial_data)
        self._transformed_initial_hyperparams_vals = transformed_data[0]
        self._transformed_initial_privacy_vals = transformed_data[1]
        self._transformed_initial_utility_vals = transformed_data[2]

        # Creates separate GPs for modeling privacy and utility.
        self._privacy_gp = self._create_privacy_gp()
        self._utility_gp = self._create_utility_gp()

        # Runs a hard-coded sanity check against data from the loaded file.
        self._print_sanity_check()

        # Sets the information for the reference (aka anti-ideal) point and its transformation.
        self._anti_ideal_point = anti_ideal_point
        transformed_anti_ideal_eps = self._transform_privacy(self._anti_ideal_point[0])
        transformed_anti_ideal_err = self._transform_utility(self._anti_ideal_point[1])
        self._transformed_anti_ideal_point = [transformed_anti_ideal_eps, transformed_anti_ideal_err]

        # Optimization domain dictates the range of each to-be-optimized-over hyperparamter. Fixed hyperparams
        # are the hyperparams that aren't optimized over.
        self._optimization_domain = optimization_domain
        self._gpflowopt_domain = self._construct_gpflowopt_domain(optimization_domain)
        self._fixed_hyperparams = fixed_hyperparams
        self._instance_options = instance_options

        # Create acquisition function, set it's anti-ideal point, and create it's optimizer.
        self._acquisition = gpflowopt.acquisition.HVProbabilityOfImprovement([self._privacy_gp, self._utility_gp])
        self._acquisition.reference = np.array([self._transformed_anti_ideal_point])
        self._acquisition_optimizer = gpflowopt.optim.StagedOptimizer([gpflowopt.optim.MCOptimizer(self._gpflowopt_domain, 5000),
                                                                       gpflowopt.optim.SciPyOptimizer(self._gpflowopt_domain)])

        # Create the Bayesian Optimizer.
        self._bayesian_optimizer = gpflowopt.BayesianOptimizer(self._gpflowopt_domain,
                                                               self._acquisition,
                                                               optimizer=self._acquisition_optimizer,
                                                               scaling=False)

        # Multiprocessing worker info
        self._num_workers = num_workers

        # Misc fields
        self._saved_hypervolumes = []
        self._all_results = []
        self._output_base_dir = output_base_dir
        self._run_name = int(time.time() * 1000)
        self._save_results = True
        self._make_results_directory()
        self._show_plot = plot_options.get('show_plot', False)
        self._plot_options = plot_options


    ##################################################################################
    # Initial data-related methods
    ##################################################################################
    @staticmethod
    def _process_replications(data):
        aggregated_points = {}
        for point in data:
            hyperparams_dict = point[0]
            frozen_hyperparams = frozenset(hyperparams_dict.items())
            if frozen_hyperparams not in aggregated_points:
                aggregated_points[frozen_hyperparams] = ([], [])
            aggregated_points[frozen_hyperparams][0].append(point[1])
            aggregated_points[frozen_hyperparams][1].append(point[2])
        
        processed_data = []
        seen_points = set()
        for point in data:
            hyperparams_dict = point[0]
            frozen_hyperparams = frozenset(hyperparams_dict.items())
            if frozen_hyperparams not in seen_points:
                seen_points.add(frozen_hyperparams)
                # compute privacy as max value
                privacy = np.max(aggregated_points[frozen_hyperparams][0])
                # compute utility as mean value
                utility = np.mean(aggregated_points[frozen_hyperparams][1])
                # add variance of utility values
                utility_variance = np.var(aggregated_points[frozen_hyperparams][1])
                # add everything back into the list of processed data
                processed_data.append((hyperparams_dict, privacy, utility, utility_variance))
        return processed_data

    def _get_initial_points(self, initial_data_options):
        num_initial_points = initial_data_options['num_points']

        # get initial data from file or generate it with random sampling
        if 'filename' in initial_data_options:
            filename = initial_data_options['filename']
            raw_data = self._load_points_from_file(filename)
        else:
            raise ValueError("Please provide path to the file full_results.pkl that contains initial data points from random sampling run.")

        # Process out the replications and extract the proper number of initial points
        processed_data = self._process_replications(raw_data)[:num_initial_points]

        # Set input hyperparams and output privacy & utility
        initial_hyperparams_dicts = np.array([instance[0] for instance in processed_data])

        initial_hyperparams_vals = np.array([np.array([instance[name] for name in self._optimizable_hyperparams_names]) for instance in initial_hyperparams_dicts])
        initial_privacy_vals = np.expand_dims(np.array([instance[1] for instance in processed_data]), axis=1)
        initial_utility_vals = 1-np.expand_dims(np.array([instance[2] for instance in processed_data]), axis=1)
        max_initial_utility_variance = np.max([instance[3] for instance in processed_data])
        return initial_hyperparams_vals, initial_privacy_vals, initial_utility_vals, max_initial_utility_variance

    def _load_points_from_file(self, filename):
        with open(filename, 'rb') as f:
            raw_data = pickle.load(f)
        return raw_data

    def _create_initial_transformations(self, initial_data):
        transformed_initial_hyperparams_vals = self._transform_initial_hyperparameters(initial_data[0])
        transformed_initial_privacy_vals = self._transform_initial_privacy(initial_data[1])
        transformed_initial_utility_vals = self._transform_utility(initial_data[2])
        return transformed_initial_hyperparams_vals, transformed_initial_privacy_vals, transformed_initial_utility_vals

    ##################################################################################
    # Hyperparameter-related methods
    ##################################################################################
    def _transform_initial_hyperparameters(self, initial_hyperparams_vals):
        self._hyperparam_normalization_factor = initial_hyperparams_vals.max(axis=0)
        return self._transform_hyperparameters(initial_hyperparams_vals)

    def _transform_hyperparameters(self, hyperparams):
        return hyperparams / self._hyperparam_normalization_factor

    def _untransform_hyperparameters(self, transformed_hyperparams):
        return transformed_hyperparams * self._hyperparam_normalization_factor

    ##################################################################################
    # Privacy-related methods
    ##################################################################################
    def _transform_initial_privacy(self, initial_privacy_vals):
        log_initial_privacy_vals = np.log(initial_privacy_vals)
        self._privacy_normalization_factor = np.asscalar(log_initial_privacy_vals.max(axis=0))
        return self._transform_privacy(initial_privacy_vals)

    def _transform_privacy(self, val):
        return np.log(val) / self._privacy_normalization_factor

    def _untransform_privacy(self, transformed_val):
        return np.exp(transformed_val * self._privacy_normalization_factor)

    def _create_privacy_gp(self):
        privacy_gp = gpflow.gpr.GPR(self._transformed_initial_hyperparams_vals.copy(),
                                    self._transformed_initial_privacy_vals.copy(),
                                    self.SafeMatern52(input_dim=5, ARD=True), name='privacy')
        privacy_gp.likelihood.variance = 0
        privacy_gp.likelihood.variance.fixed = True
        privacy_gp.optimize()
        return privacy_gp

    def predict_privacy(self, hyperparams, transform_hyperparams=False, untransform_prediction=True):
        gp_input = self._transform_hyperparameters(hyperparams) if transform_hyperparams else hyperparams
        pred = self._privacy_gp.predict_f(gp_input)[0]
        if untransform_prediction:
            pred = self._untransform_privacy(pred)
        return pred

    ##################################################################################
    # Utility-related methods
    ##################################################################################
    def _transform_initial_utility(self, initial_utility_vals):
        return self._transform_utility(initial_utility_vals)

    def _transform_utility(self, val):
        return sp.special.logit(val)

    def _untransform_utility(self, transformed_val):
        return sp.special.expit(transformed_val)

    def _create_utility_gp(self):
        utility_gp = gpflow.gpr.GPR(self._transformed_initial_hyperparams_vals.copy(),
                                    self._transformed_initial_utility_vals.copy(),
                                    self.SafeMatern52(input_dim=5, ARD=True), name='utility')
        utility_gp.likelihood.variance = 0.01
        utility_gp.optimize()
        return utility_gp

    def predict_utility(self, hyperparams, transform_hyperparams=False, untransform_prediction=True):
        gp_input = self._transform_hyperparameters(hyperparams) if transform_hyperparams else hyperparams
        pred = self._utility_gp.predict_f(gp_input)[0]
        if untransform_prediction:
            pred = self._untransform_utility(pred)
        return pred

    ##################################################################################
    # Multi-objective optimization-related methods
    ##################################################################################
    def predict(self, hyperparams, transform_hyperparams=False, untransform_prediction=True):
        privacy_pred = self.predict_privacy(hyperparams, transform_hyperparams, untransform_prediction)
        utility_pred = self.predict_utility(hyperparams, transform_hyperparams, untransform_prediction)
        return privacy_pred, utility_pred

    def _construct_gpflowopt_domain(self, optimization_domain):
        gpflowopt_domain = None
        for idx, name in enumerate(self._optimizable_hyperparams_names):
            transformed_domain = optimization_domain[name] / self._hyperparam_normalization_factor[idx]
            new_dim = gpflowopt.domain.ContinuousParameter(name, *transformed_domain)
            if not gpflowopt_domain:
                gpflowopt_domain = new_dim
            else:
                gpflowopt_domain += new_dim
        return gpflowopt_domain

    def get_transformed_hypervolume(self):
        hv = self._acquisition.pareto.hypervolume(self._transformed_anti_ideal_point)
        return hv

    def get_untransformed_hypervolume(self):
        # Get untransformed pf points
        pf, dom = gpflowopt.pareto.non_dominated_sort(self._acquisition.data[1])
        pf_eps = self._untransform_privacy(pf[:, 0])
        pf_err = self._untransform_utility(pf[:, 1])
        # Sort both lists by increasing x-axis (eps)
        pf_eps, pf_err = (list(t) for t in zip(*sorted(zip(pf_eps, pf_err))))

        # TODO: Do this before untransforming (in case the untransform doesn't preserve ordering).
        # Beginning with the right-most point on the pf, discard out of both lists if it is not in the region-of-interest
        for i in range(len(pf_eps)-1, -1):
            if pf_eps > self._anti_ideal_point[0] or pf_err > self._anti_ideal_point[1]:
                del pf_eps[i]
                del pf_err[i]

        hv = 0
        if len(pf_eps) > 0:
            for i in range(len(pf_eps)-1, 0):
                # Accumulate the rectangle formed by: 1) current point to the anti-ideal privacy point, 2) current point up to the next point.
                hv += ((self._anti_ideal_point[0] - pf_eps[i]) * (pf_err[i-1] - pf_err[i]))
            # Final point on the pf has rectangle formed by: 1) current point to the anti-ideal privacy point, 2) current point to the anti-ideal error point.
            hv += ((self._anti_ideal_point[0] - pf_eps[0]) * (self._anti_ideal_point[1] - pf_err[0]))
        return hv

    def run(self):
        self._previous_hyperparams = None
        self._optimization_point_number = 0

        optimization_result = self._bayesian_optimizer.optimize([self._optimization_handler], n_iter=self._num_instances)

        # Post final run tasks
        hypervolume = self.get_untransformed_hypervolume()
        self._saved_hypervolumes.append(hypervolume)
        print("Hypervolume is: {}".format(hypervolume))
        self._estimate_last_point()
        self._plot_and_save()

    ##################################################################################
    # Methods relating to the expensive function being optimized over.
    ##################################################################################
    def _convert_model_input_to_problem_instance_input(self, transformed_hyperparams_arr):
        hyperparams_arr = self._untransform_hyperparameters(transformed_hyperparams_arr)
        hyperparams_dict = self._fixed_hyperparams.copy()
        # TODO: make better
        for idx, name in enumerate(self._optimizable_hyperparams_names):
            val = hyperparams_arr[0][idx]
            if name in ['epochs', 'lot_size']:
                val = int(round(val))
            hyperparams_dict[name] = val
        return hyperparams_dict

    def _run_pre_iteration_tasks(self, transformed_hyperparams):
        # Print and store hypervolume from previous run
        hypervolume = self.get_untransformed_hypervolume()
        self._saved_hypervolumes.append(hypervolume)
        print("Hypervolume is: {}".format(hypervolume))

        # Run check to ensure previously acquired point was learned properly, then save current hyperparams for next run.
        self._estimate_last_point()
        self._previous_hyperparams = transformed_hyperparams

        # Untransform hyperparams and put into our expected format (a dictionary of hyperparam names to vals).
        hyperparams_dict = self._convert_model_input_to_problem_instance_input(transformed_hyperparams)

        # Predict privacy and utility before training model.
        estimated_privacy, estimated_utility = self.predict(transformed_hyperparams, transform_hyperparams=False)
        print("curr point: {}".format(hyperparams_dict))
        print("curr pts estimated (eps, err): {}".format((estimated_privacy, estimated_utility)))

        self._plot_and_save()

        self._optimization_point_number += 1

        save_object('{}/optimization-hyperparameters'.format(self._results_dir),
                    hyperparams_dict,
                    'round-{}'.format(self._optimization_point_number))
        
        return hyperparams_dict

    def _run_post_iteration_tasks(self, eps, acc):
        self._print_sanity_check()

        transformed_eps = self._transform_privacy(eps)
        transformed_err = self._transform_utility(1-acc)
        
        return transformed_eps, transformed_err

    def _run_replications(self, hyperparams_dict, options):
        print("Beginning run of {} replications using {} workers.".format(self._num_replications, self._num_workers))

        # Create a list of inputs for the to-be-spawned process
        training_instances_inputs = []
        for i in range(self._num_replications):
            process_hyperparams = hyperparams_dict.copy()
            process_options = options.copy()
            # Compute privacy only for the first instance of its kind, not for any subsequent replications. Determine
            # this simply using the ordering of the instances in the hyperparam_instances list
            if i > 0:
                process_options['accumulate_privacy'] = False
                process_options['verbose'] = False
            training_instance_input = {'hyperparams': process_hyperparams,
                                       'options': process_options,
                                       'model_class': self._model_class}
            training_instances_inputs.append(training_instance_input)

        # Begin multiprocessing across the instances
        self._id_queue = mp.Manager().Queue()
        for pid in range(self._num_workers):
            self._id_queue.put(pid)
        with Pool(self._num_workers, self._process_id_init, (self._id_queue,)) as pool:
            results = pool.map_async(self._run_single_instance, training_instances_inputs, chunksize=1)
            # while not results.ready():
            #     completed_instances = self._num_replications - results._number_left
            #     print("Status: {} of {} complete.".format(completed_instances, self._num_replications), end='\r')
            #     time.sleep(1)
            self._all_results += results.get()

        # process results
        aggregated_privacy = np.max([result[0] for result in results.get()])
        aggregated_utility = np.mean([result[1] for result in results.get()])
        return aggregated_privacy, aggregated_utility

    # Given some hyperparameters and options, runs a single instance on its own compute device
    @staticmethod
    def _run_single_instance(inputs):

        hyperparams = inputs['hyperparams']
        options = inputs['options']
        model_class = inputs['model_class']

        global process_id
        options.update({'device_id': process_id})
        # Ensures the process is running on the CPU identified by process_id.
        proc = psutil.Process()
        try:
            proc.cpu_affinity([process_id])
        except:
            pass  # doesn't work on Macs...but Macs seem to handle CPU assignment fine without this, so just continue.

        # Change this process' random seed
        custom_seed = int(time.time() + hash(frozenset(hyperparams.items())) + 547*float(process_id)) # don't ask
        set_random_seed(custom_seed)

        print("Worker {} evaluating hyperparameters {}.".format(process_id, hyperparams))

        # Run instance with random hyperparams
        priv, acc = -1, -1
        try:
            instance = model_class(hyperparams, options)
            priv, acc = instance.run()
            gc.collect()  # necessary to avoid out-of-memory errors
            if 'accumulate_privacy' in options and options['accumulate_privacy'] == False:
                priv = -2
        # If an instance crashes, just return (-1,-1) as a failure code
        except Exception as e:
            print("Worker {} running with hyperparameters \n{}\n threw the following exception:\n{}".format(process_id, hyperparams, e))
        finally:
            print("Worker {} returning result: ({}, {})".format(process_id, priv, acc))
            return priv, acc

    # Gets the unique process id (to be used to assign a unique compute device)
    @staticmethod
    def _process_id_init(queue):
        global process_id
        process_id = queue.get()

    def _optimization_handler(self, transformed_hyperparams):
        # GPFlowOpt discards output on Windows; the following lines correct this.
        import sys
        sys.stdout = sys.__stdout__

        # Do any computation before testing a point on the objective function.
        hyperparams_dict = self._run_pre_iteration_tasks(transformed_hyperparams)

        options = self._instance_options
        options['output_base_dir'] = self._results_dir

        eps, acc = self._run_replications(hyperparams_dict, options)
        print("actual (eps, err): {}".format((eps, 1-acc)))

        # Do any computation after testing a point on the objective function before updating the model.
        transformed_eps, transformed_err = self._run_post_iteration_tasks(eps, acc)

        return np.array([(transformed_eps, transformed_err)])

    def _estimate_last_point(self):
        if self._previous_hyperparams is not None:
            estimated_privacy, estimated_utility = self.predict(self._previous_hyperparams, transform_hyperparams=False)
            print("#######################################################################################")
            print("Checking last point acquired in optimization loop.")
            print("estimated (eps, err) of point: {}".format((estimated_privacy, estimated_utility)))
            print("#######################################################################################")

    ##################################################################################
    # Miscellaneous code.
    ##################################################################################
    def _plot_and_save(self):
        self._plot_empirical_pareto_frontier(suffix=str(self._optimization_point_number))

        # save_object('{}/saved_gp_models'.format(self._results_dir),
        #             [self._privacy_gp, self._utility_gp],
        #             'round-{}'.format(self._optimization_point_number),
        #             save_text=False)

        save_object('{}/saved_optimization_state'.format(self._results_dir),
                    self,
                    'round-{}'.format(self._optimization_point_number),
                    save_text=False)

    def _plot_empirical_pareto_frontier(self, suffix=''):
        pf, dom = gpflowopt.pareto.non_dominated_sort(self._acquisition.data[1])
        pf_eps = self._untransform_privacy(pf[:, 0])
        pf_err = self._untransform_utility(pf[:, 1])
        # Sort both lists by increasing x-axis (eps)
        pf_eps, pf_err = (list(t) for t in zip(*sorted(zip(pf_eps, pf_err))))

        all_eps = self._untransform_privacy(self._acquisition.data[1][:, 0])
        all_err = self._untransform_utility(self._acquisition.data[1][:, 1])

        plt.figure(figsize=(9, 7))

        # Plot all points, color by dominance
        plt.scatter(all_eps, all_err, c=dom, marker='.')

        # Mark all points on the pf with red X and build a stepwise dominance line
        plt.scatter(pf_eps, pf_err, color='r', marker='x')
        boundary_pt = ((pf_eps[0], 2*self._anti_ideal_point[1]), (2*self._anti_ideal_point[0], pf_eps[-1]))
        pf_eps.insert(0, boundary_pt[0][0])
        pf_err.insert(0, boundary_pt[0][1])
        pf_eps.insert(len(pf_eps), boundary_pt[1][0])
        pf_err.insert(len(pf_err), boundary_pt[1][1])
        plt.step(pf_eps, pf_err, 'k-', where='post')

        # Plot options
        self._set_plot_properties(plt, self._plot_options)
        if self._save_results:
            plots_dir = '{}/pareto-front-plots'.format(self._results_dir)
            os.makedirs(plots_dir, exist_ok=True)
            plt.savefig('{}/round-{}.png'.format(plots_dir, suffix))
        if self._show_plot:
            # TODO: resolve blocking issue when showing plot on Windows
            plt.show()

    def _set_plot_properties(self, plt, plot_options):
        title = plot_options.get('title', 'Empirical Pareto Frontier')
        xlabel = plot_options.get('xlabel', 'epsilon')
        ylabel = plot_options.get('ylabel', 'classification error')
        plt.title(title, size=20)
        plt.xlabel(xlabel, size=18)
        plt.ylabel(ylabel, size=18)
        
        left = plot_options.get('left', 0)
        right = plot_options.get('right', self._anti_ideal_point[0])
        bottom = plot_options.get('bottom', 0)
        top = plot_options.get('top', self._anti_ideal_point[1])
        plt.xlim(left=left, right=right)
        plt.ylim(bottom=bottom, top=top)

    def _print_sanity_check(self):
        print("#######################################################################################")
        print("Running GP sanity check on a training point.")

        # first point in loaded data
        first_point = np.expand_dims(self._initial_hyperparams_vals[0], axis=0)

        actual_privacy = self._initial_privacy_vals[0]
        actual_utility = self._initial_utility_vals[0]
        print("actual (eps, err): {}".format((actual_privacy, actual_utility)))

        # gp estimate of first point
        estimated_privacy, estimated_utility = self.predict(first_point, transform_hyperparams=True)
        print("estimated (eps, err): {}".format((estimated_privacy, estimated_utility)))

        # perturbed first point in loaded data (lot size increase by 1)
        almost_first_point = np.copy(first_point)
        almost_first_point[0][self._optimizable_hyperparams_names.index('lot_size')] += 1
        estimated_privacy, estimated_utility = self.predict(almost_first_point, transform_hyperparams=True)
        print("estimated (eps, err) of perturbed: {}".format((estimated_privacy, estimated_utility)))

        print("#######################################################################################")

    def _make_results_directory(self):
        if not self._output_base_dir:
            self._results_dir = 'results/hvpoi_results/{}/{}'.format(self._problem_name, self._run_name)
        else:
            self._results_dir = '{}/hvpoi_results/{}'.format(self._output_base_dir, self._run_name)
        os.makedirs(self._results_dir, exist_ok=True)

    # Matern52 kernel isn't numerically safe, so this work-around is necessary.
    class SafeMatern52(gpflow.kernels.Matern52):
        def euclid_dist(self, X, X2):
            r2 = self.square_dist(X, X2)
            return tf.sqrt(r2 + 1e-7)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import mxnet as mx
from mxnet import nd

from dpareto.optimizers.dp_optimizer import DpOptimizer

class DpAdam(DpOptimizer):
    def __init__(self, hyperparams, net, params, loss_func, model_ctx, accountant):
        super(DpAdam, self).__init__(hyperparams, net, params, loss_func, model_ctx, accountant)

        # Compute scale of Gaussian noise to add
        self._hyperparams['sigma'] = hyperparams['z'] * (2 * hyperparams['l2_clipping_bound'] / hyperparams['lot_size'])

        # Initialize 1st and 2nd moment vectors
        self._m = {}
        self._v = {}
        for param_name, param in self._params.items():
            self._m[param_name] = nd.zeros_like(param)
            self._v[param_name] = nd.zeros_like(param)

    def _update_params(self, accumulated_grads):
        # scale gradients by lot size, add noise, and update the parameters
        for param_name, param in self._params.items():
            # average the clipped gradients and then add noise to each averaged gradient
            param_grad_update = (accumulated_grads[param_name] / self._hyperparams['lot_size']) + \
                                mx.random.normal(0, self._hyperparams['sigma'], param.shape, ctx=self._model_ctx)

            # update biased first moment estimate
            self._m[param_name] = self._hyperparams['beta_1'] * self._m[param_name] + (1 - self._hyperparams['beta_1']) * param_grad_update

            # update biased second raw moment estimate
            self._v[param_name] = self._hyperparams['beta_2'] * self._v[param_name] + (1 - self._hyperparams['beta_2']) * nd.square(param_grad_update)

            # compute bias-corrected first moment estimate
            m_hat = self._m[param_name] / (1 - nd.power(self._hyperparams['beta_1'], self._step + 1))

            # compute bias-corrected second raw moment estimate
            v_hat = self._v[param_name] / (1 - nd.power(self._hyperparams['beta_2'], self._step + 1))

            # update params with ADAM
            param[:] = param - self._hyperparams['lr'] * m_hat / (nd.sqrt(v_hat) + 1e-8)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

from mxnet.gluon.data import sampler
import numpy as np


class LotSampler(sampler.BatchSampler):
    def __init__(self, lot_size, data_size):
        self._lot_size = lot_size
        self._data_size = data_size
        super().__init__(self.RandomBatchSampler(lot_size, data_size), lot_size)

    class RandomBatchSampler(sampler.Sampler):
        """Samples a subset of elements randomly without replacement from the full dataset.
        Parameters
        ----------
        num_total : int
            Number of elements total to sample from.
        num_sample : int
            Number of elements to be sampled.
        """
        def __init__(self, num_sample, num_total):
            self._num_sample = num_sample
            self._num_total = num_total

        def __iter__(self):
            indices = np.arange(self._num_total)
            np.random.shuffle(indices)
            indices = indices[:self._num_sample]
            return iter(indices)

        def __len__(self):
            return self._num_sample



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import os
import pickle

import autodp as adp
import dpareto.data.adult.importer as importer
from dpareto.utils.object_io import save_object


# TODO sample_fraction is generally not available as input to the algorithm!
def compute_epsilon(hyperparams, num_training_examples=None):

    if 'sample_fraction' in hyperparams:
        sample_fraction = hyperparams['sample_fraction']
        num_iterations = hyperparams['epochs'] * round(hyperparams['lot_size'] / sample_fraction)
    elif num_training_examples is not None:
        num_iterations = hyperparams['epochs'] * round(num_training_examples / hyperparams['lot_size'])
        sample_fraction = hyperparams['lot_size'] / num_training_examples
    else:
        raise ValueError('[compute_epsilon] Missing both num_training_examples and sample_fraction')

    accountant = adp.rdp_acct.anaRDPacct()
    cgf_func = lambda x: adp.rdp_bank.RDP_gaussian({'sigma': hyperparams['z']}, x)
    for i in range(num_iterations):
        accountant.compose_subsampled_mechanism(cgf_func, sample_fraction)
    return accountant.get_eps(hyperparams['fixed_delta'])


def get_adult_training_size():
    Xtrain, Ytrain, Xtest, Ytest = importer.import_adult_dataset(None)
    return len(Ytrain)
    # return 30956


def get_mnist_training_size():
    return 60000


def convert_from_pydiffpriv_to_autodp(num_training_examples, results_directory, results_name='full_results'):
    # get results
    original_results_base = '{}/{}'.format(results_directory, results_name)
    with open('{}.pkl'.format(original_results_base), 'rb') as f:
        original_results = pickle.load(f)

    # for each result, if privacy was computed, recompute privacy with autodp
    auto_dp_results = []
    for (hyperparams, pydiffpriv_privacy, utility) in original_results:
        if pydiffpriv_privacy > 0:
            autodp_privacy = compute_epsilon(hyperparams, num_training_examples)
        else:
            autodp_privacy = pydiffpriv_privacy
        # store result in same format as before
        auto_dp_results.append((hyperparams, autodp_privacy, utility))

    # rename old results to results_name+'-pydiffpriv' if it doesn't already exist
    if not os.path.isfile('{}-pydiffpriv.pkl'.format(original_results_base)):
        os.rename('{}.pkl'.format(original_results_base), '{}-pydiffpriv.pkl'.format(original_results_base))
        os.rename('{}.txt'.format(original_results_base), '{}-pydiffpriv.txt'.format(original_results_base))
    else:
        print('Old pydiffpriv files already exist. Continuing anyways.')

    # save new results to results_name
    save_object(results_directory, auto_dp_results, results_name)


if __name__ == "__main__":
    # Debug a problematic instance for autoDP
    Ntrain = get_adult_training_size()
    hp = {'fixed_delta': 1e-05, 'epochs': 2, 'lot_size': 8, 'lr': 0.006699102447124356, 'l2_clipping_bound': 2.909185033175768, 'z': 12.67021387220363}
    eps = compute_epsilon(hp, Ntrain)
    print(eps)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import pickle
import mxnet as mx
from mxnet import nd
import urllib.request
import shutil
import ssl


data_dir = "dpareto/data/adult"

data_urls = {}
with open(f"{data_dir}/urls.txt") as f:
    for line in f:
       (dataset, url) = line.split(": ")
       data_urls[dataset] = url

print("Downloading...")
context = ssl._create_unverified_context()  # Bypasses failing HTTPS verification for the dataset site
training_data = urllib.request.urlopen(data_urls["training"], context=context).read().decode('utf-8')
testing_data = urllib.request.urlopen(data_urls["testing"], context=context).read().decode('utf-8')
print("downloaded.")

def process_data(raw_data):
    data_ctx = mx.cpu()
    lines = raw_data.splitlines()
    num_examples = len(lines)
    num_features = 123
    X = nd.zeros((num_examples, num_features), ctx=data_ctx)
    Y = nd.zeros((num_examples, 1), ctx=data_ctx)
    for i, line in enumerate(lines):
        tokens = line.split()
        label = (int(tokens[0]) + 1) / 2  # Change label from {-1,1} to {0,1}
        Y[i] = label
        for token in tokens[1:]:
            index = int(token[:-2]) - 1
            X[i, index] = 1
    return X, Y

print("Processing...")
Xtrain, Ytrain = process_data(training_data)
Xtest, Ytest = process_data(testing_data)
print("processed.")

print("Pickling...")
pickle.dump( (Xtrain, Ytrain), open( f"{data_dir}/a1a.train.p", "wb" ) )
pickle.dump( (Xtest, Ytest), open( f"{data_dir}/a1a.test.p", "wb" ) )
print("pickled.")



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import mxnet as mx
from mxnet import nd, gluon
import numpy as np

import dpareto.data.adult.importer as importer
from dpareto.models.dp_feedforward_net import DpFeedforwardNet
from dpareto.utils.lot_sampler import LotSampler


class AdultBase(DpFeedforwardNet):
    def __init__(self, hyperparams, options={}):
        super(AdultBase, self).__init__(hyperparams, options)

    def _load_data(self):
        xTrain, yTrain, xTest, yTest = importer.import_adult_dataset(self._data_ctx)

        num_training_examples = len(yTrain)
        num_testing_examples = len(yTest)

        if self._verbose:
            print("Loading...")
        train_data = gluon.data.ArrayDataset(xTrain, yTrain)
        train_data_lot_iterator = gluon.data.DataLoader(train_data,
                                                        batch_sampler=LotSampler(self._lot_size, num_training_examples))
        train_data_eval_iterator = gluon.data.DataLoader(train_data, batch_size=self._lot_size)
        test_data = gluon.data.DataLoader(gluon.data.ArrayDataset(xTest, yTest),
                                          batch_size=self._lot_size, shuffle=True)
        if self._verbose:
            print("done loading.")

        return num_training_examples, num_testing_examples, train_data_lot_iterator, train_data_eval_iterator, test_data

    def _evaluate_accuracy(self, data_iterator):
        num_correct = 0
        total = 0
        for i, (data, label) in enumerate(data_iterator):
            data = data.as_in_context(self._model_ctx).reshape((-1, 1, self._input_layer))
            label = label.as_in_context(self._model_ctx)

            output = self._net(data, self._params, training_mode=False)
            prediction = output > 0.0

            num_correct += nd.sum(prediction == label)
            total += len(label)

        return num_correct.asscalar() / total



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import pickle
import os


def save_object(output_dir, obj, name, save_pickle=True, save_text=True):
    os.makedirs(output_dir, exist_ok=True)

    if save_pickle:
        # Save pickle of object
        with open('{}/{}.pkl'.format(output_dir, name), 'wb') as f:
            pickle.dump(obj, f, pickle.HIGHEST_PROTOCOL)

    if save_text:
        # Save text of object
        with open('{}/{}.txt'.format(output_dir, name), 'w') as f:
            if type(obj) in (list, tuple):
                obj_string = ''
                for item in obj:
                    obj_string += str(item) + '\n'
            elif type(obj) is dict:
                obj_string = ''
                for key, value in obj.items():
                    obj_string += str(key) + '\n - ' + str(value) + '\n'
            else:
                obj_string = str(obj)
            f.write(obj_string)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import argparse
import gc
import math
import multiprocessing as mp
from multiprocessing import Pool
import random
import time
import os
import psutil

from dpareto.utils.mxnet_scripts import get_gpu_count
from dpareto.utils.object_io import save_object
from dpareto.utils.random_seed_setter import set_random_seed


class RandomSamplingHarness:
    def __init__(self, model_class, problem_name, hyperparam_distributions, instance_options={}, num_instances=1, num_replications=1, num_workers=1, output_base_dir=''):
        mp.set_start_method('spawn', force=True)  # workaround for weird matplotlib+multiprocessing compatibility issue

        self._model_class = model_class
        self._problem_name = problem_name

        self._hyperparam_distributions = hyperparam_distributions
        self._instance_options = instance_options
        self._num_instances = num_instances
        self._num_replications = num_replications
        self._num_workers = num_workers

        # Overwrite hardcoded options with command-line-supplied options
        self._apply_args()

        # Multiprocessing worker info
        self._id_queue = mp.Manager().Queue()
        for pid in range(self._num_workers):
            self._id_queue.put(pid)

        # Output info
        self._output_base_dir = output_base_dir
        self._run_name = int(time.time() * 1000)
        self._make_results_directory()

    def _apply_args(self):
        parser = argparse.ArgumentParser()
        parser.add_argument('--device')
        parser.add_argument('--workers')
        parser.add_argument('--instances')
        args = parser.parse_args()

        # Set device if provided
        if args.device is not None:
            if args.device == 'gpu':
                self._instance_options['use_gpu'] = True
            else:
                self._instance_options['use_gpu'] = False
        # Otherwise, default to CPU if nothing is provided in instance_options
        else:
            if 'use_gpu' not in self._instance_options:
                # if not specified or if 'cpu'
                self._instance_options['use_gpu'] = False

        # Set number of instances if provided.
        if args.instances is not None:
            self._num_instances = int(args.instances)

        # Set number of workers if provided. An explicit number or 'max' can be provided.
        if args.workers is not None:
            if args.workers == 'max':
                # get number of CPU cores if use_gpu is False
                if not self._instance_options['use_gpu']:
                    self._num_workers = mp.cpu_count()
                else:  # get number of GPU cores
                    self._num_workers = get_gpu_count()
                    if self._num_workers == 0:
                        print("No GPUs available. Exiting.")
                        exit()
            else:
                self._num_workers = int(args.workers)

    # Main runner of the random sampling framework
    def run(self):
        self._generate_random_hyperparameter_instances()
        self._run_all_instances()
        self._store_results()
        return self._results

    # Generates num_instances number of hyperparameter setting instances
    def _generate_random_hyperparameter_instances(self):
        print("Generating random hyperparameters...", end='')
        self._hyperparam_instances = []
        for i in range(self._num_instances):
            # Generate a single random set of hyperparameters
            hyperparam_instance = {}
            for name, hyperparam_distribution in self._hyperparam_distributions.items():
                hyperparam_instance[name] = self._random_sample(hyperparam_distribution)
            for j in range(self._num_replications):
                self._hyperparam_instances.append(hyperparam_instance)
        save_object(self._results_dir, self._hyperparam_instances, 'hyperparams')
        print("done. Generated and saved.")

    # From the instances (and their replications) of random hyperparameters generated, coordinates the parallel runs
    # of all the training instances corresponding to each hyperparameter instance
    def _run_all_instances(self):
        print("Beginning {} runs distributed across {} workers.".format(len(self._hyperparam_instances), self._num_workers))

        # Create a list of inputs for the to-be-spawned process
        training_instances_inputs = []
        for idx, hyperparam_instance in enumerate(self._hyperparam_instances):
            single_instance_hyperparams = hyperparam_instance.copy()
            single_instance_options = self._instance_options.copy()
            # Compute privacy only for the first instance of its kind, not for any subsequent replications. Determine
            # this simply using the ordering of the instances in the hyperparam_instances list
            if idx > 0 and self._hyperparam_instances[idx] == self._hyperparam_instances[idx-1]:
                single_instance_options['accumulate_privacy'] = False
            training_instance_input = {'hyperparams': single_instance_hyperparams,
                                       'options': single_instance_options,
                                       'model_class': self._model_class}
            training_instances_inputs.append(training_instance_input)

        # Begin multiprocessing across the instances
        with Pool(self._num_workers, self._process_id_init, (self._id_queue,)) as pool:
            results = pool.map_async(self._run_single_instance, training_instances_inputs, chunksize=1)
            while not results.ready():
                completed_instances = len(self._hyperparam_instances) - results._number_left
                print("Status: {} of {} complete.".format(completed_instances, len(self._hyperparam_instances)), end='\r')
                time.sleep(1)
            self._results = results.get()

    # Given some hyperparameters and options, runs a single instance on its own compute device
    @staticmethod
    def _run_single_instance(inputs):
        hyperparams = inputs['hyperparams']
        options = inputs['options']
        model_class = inputs['model_class']

        global process_id
        options.update({'device_id': process_id})
        # Ensures the process is running on the CPU identified by process_id.
        proc = psutil.Process()
        try:
            proc.cpu_affinity([process_id])
        except:
            pass  # doesn't work on Macs...but Macs seem to handle CPU assignment fine without this, so just continue.

        # Change this process' random seed
        custom_seed = int(time.time() + hash(frozenset(hyperparams.items())) + 547*float(process_id)) # don't ask
        set_random_seed(custom_seed)

        print("Worker {} evaluating hyperparameters {}.".format(process_id, hyperparams))

        # Run instance with random hyperparams
        priv, acc = -1, -1
        try:
            priv, acc = model_class(hyperparams, options).run()
            gc.collect()  # necessary to avoid out-of-memory errors
            if 'accumulate_privacy' in options and options['accumulate_privacy'] == False:
                priv = -2
        # If an instance crashes, just return (-1,-1) as a failure code
        except Exception as e:
            print("Worker {} running with hyperparameters \n{}\n threw the following exception:\n{}".format(process_id, hyperparams, e))
        finally:
            print("Worker {} returning result: ({}, {})".format(process_id, priv, acc))
            return priv, acc

    # Gets the unique process id (to be used to assign a unique compute device)
    @staticmethod
    def _process_id_init(queue):
        global process_id
        process_id = queue.get()

    # Handles the random sampling, given a range and some additional properties of the hyperparameter.
    @staticmethod
    def _random_sample(distribution):
        distribution_type = distribution.get('type')

        reject_less_than = distribution.get('reject_less_than', False)
        reject_greater_than = distribution.get('reject_greater_than', False)
        if not reject_less_than:
            reject_less_than = -math.inf
        if not reject_greater_than:
            reject_greater_than = math.inf

        # Sample until number is within an acceptable range
        while True:
            if distribution_type == 'uniform':
                a = min(distribution['params'])
                b = max(distribution['params'])
                sample = random.uniform(a, b)
            elif distribution_type == 'normal':
                # sample from normal dist
                mu = distribution['params'][0]
                sigma = distribution['params'][1]
                sample = random.normalvariate(mu, sigma)
            elif distribution_type == 'exponential':
                # sample from exponential dist
                lambd = distribution['params'][0]
                shift = distribution['params'][1]
                sample = shift + random.expovariate(lambd)
            elif distribution_type == 'deterministic':
                sample = distribution['value']
            else:
                raise NotImplementedError("Must specify a supported sampling distribution.")

            if reject_less_than < sample < reject_greater_than:
                break

        # Optional: Round to nearest int
        if distribution.get('round_to_int', False):
            sample = int(round(sample))

        # Optional: Clip to bounds
        lower_bound = distribution.get('lower_bound', False)
        if lower_bound:
            sample = max(sample, lower_bound)
        upper_bound = distribution.get('upper_bound', False)
        if upper_bound:
            sample = min(sample, upper_bound)

        return sample

    def _store_results(self):
        full_results = [(self._hyperparam_instances[i], self._results[i][0], self._results[i][1])
                        for i in range(len(self._hyperparam_instances))]
        save_object(self._results_dir, full_results, 'full_results')

    def _make_results_directory(self):
        if not self._output_base_dir:
            self._results_dir = 'results/random_sampling_results/{}/{}'.format(self._problem_name, self._run_name)
        else:
            self._results_dir = '{}/random_sampling_results/{}'.format(self._output_base_dir, self._run_name)
        os.makedirs(self._results_dir, exist_ok=True)
        self._instance_options['output_base_dir'] = self._results_dir



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import math
import mxnet as mx
from mxnet import nd, autograd
# from pydiffpriv import cgfbank
from autodp import rdp_bank

# This is to shut pyDiffPriv up about small lambdas. TODO: Change in local pyDiffPriv code or suggest an update.
import warnings
warnings.filterwarnings("ignore")


class DpOptimizer:
    def __init__(self, hyperparams, net, params, loss_func, model_ctx, accountant):
        self._hyperparams = hyperparams

        # Store network and parameter info
        self._net = net
        self._params = params
        self._loss_func = loss_func
        self._model_ctx = model_ctx

        # Store privacy info
        self._accountant = accountant
        # self._cgf_func = lambda x: cgfbank.CGF_gaussian({'sigma': self._hyperparams['z']}, x)
        self._cgf_func = lambda x: rdp_bank.RDP_gaussian({'sigma': self._hyperparams['z']}, x)

        # Keep track of the number of steps (i.e., # of updates to the params vector)
        self._step = 0

        # Use a batch_size that fits in GPU memory
        self._batch_size = self._compute_good_batch_size()

    # Perform a single optimization step
    def step(self, data, labels, accumulate_privacy=True, run_training=True):
        lot_mean_loss = 0
        if run_training:
            # perform minimization step and update the parameters
            lot_mean_loss = self._minimize(data, labels)

        # update current amount of privacy budget consumed
        if accumulate_privacy:
            self._accountant.compose_subsampled_mechanism(self._cgf_func, self._hyperparams['sample_fraction'])

        self._step += 1

        return lot_mean_loss

    # Minimization of lot loss with batch processing
    def _minimize(self, data, labels):
        lot_loss = 0

        # Create storage for batches of summed gradients
        accumulated_grads = {}
        for param_name, param in self._params.items():
            accumulated_grads[param_name] = nd.zeros_like(param)

        for start_idx in range(0, self._hyperparams['lot_size'], self._batch_size):
            end_idx = min(self._hyperparams['lot_size'], start_idx + self._batch_size)
            batch_data = nd.slice_axis(data, axis=0, begin=start_idx, end=end_idx)
            batch_labels = nd.slice_axis(labels, axis=0, begin=start_idx, end=end_idx)
            # compute sum of clipped gradients for this batch of this lot
            lot_loss += self._accumulate_batch_gradients(batch_data, batch_labels, accumulated_grads)
            # then wait for computation to finish so that memory can be cleaned up before next batch
            nd.waitall()

        # use the computed gradients to update the parameters
        self._update_params(accumulated_grads)

        # block here, since the next step will depend on this result
        return lot_loss.asscalar() / self._hyperparams['lot_size']

    # Optimizer-specific parameter update method
    def _update_params(self, accumulated_grads):
        raise NotImplementedError("Method not implemented.")

    # Computes gradients for a single batch, adds them into the gradient accumulator
    def _accumulate_batch_gradients(self, data, labels, accumulated_grads):
        num_in_batch = data.shape[0]

        batch_params = self._get_batch_params(num_in_batch)
        with autograd.record():
            output = self._net(data, batch_params, training_mode=True)
            loss = self._loss_func(output, labels)
        loss.backward()

        px_clipping_factors = self._compute_px_clipping_factors(batch_params, num_in_batch)

        for param_name, param in self._params.items():
            batch_param_grad = batch_params[param_name].grad
            # clip the gradients of each example using the computed clipping factors
            clipped_batch_grads = self._clip_px_gradients(batch_param_grad, px_clipping_factors)
            # sum the gradients up along the batch axis, then store the result in the accumulator
            accumulated_grads[param_name] += nd.sum(clipped_batch_grads, axis=0)

        return nd.sum(loss)

    # Manually broadcast params by adding a dimension of lot_size size, that we can get per-example gradients
    def _get_batch_params(self, num_in_batch):
        batch_params = {}
        for param_name, param in self._params.items():
            batch_param = param.broadcast_to((num_in_batch,) + param.shape)
            batch_param.attach_grad()
            batch_params[param_name] = batch_param
        return batch_params

    # Compute how much the gradients of each example should be clipped
    def _compute_px_clipping_factors(self, batch_params, num_in_batch):
        px_norms = self._compute_px_gradient_norms(batch_params, num_in_batch)
        l2_rescales = self._hyperparams['l2_clipping_bound'] / (px_norms + 1e-8)  # tiny additive term to prevent div_by_0
        one_vs_rescale = nd.stack(nd.ones_like(px_norms, ctx=self._model_ctx), l2_rescales, axis=1)
        return nd.min(one_vs_rescale, axis=0, exclude=True)

    # Compute norms of gradients for each example in the batch
    def _compute_px_gradient_norms(self, batch_params, num_in_batch):
        batch_norms = nd.zeros(shape=(num_in_batch), ctx=self._model_ctx)
        for param_name, batch_param in batch_params.items():
            batch_norms += nd.sum(nd.square(batch_param.grad), axis=0, exclude=True)
        return nd.sqrt(batch_norms)

    def _clip_px_gradients(self, batch_grads, px_clipping_factors):
        # hacky workaround for not knowing how to multiply a (b,) shape array with a (b, x) or (b, x, y) shape array
        expanded_batch_clipping_factors = nd.expand_dims(px_clipping_factors, 1)
        if len(batch_grads.shape) == 3:
            expanded_batch_clipping_factors = nd.expand_dims(expanded_batch_clipping_factors, 1)
        return nd.multiply(batch_grads, expanded_batch_clipping_factors)

    # Rough estimate for the largest batch size <= lot_size that will fit in 16GB of GPU memory and yield fairly even-size batches
    def _compute_good_batch_size(self):
        GB_LIMIT = 16.
        FUDGE_FACTOR = 2.71
        PARAM_SIZE_IN_GB = 64. / (8 * 1024 * 1024 * 1024)
        num_params = sum([param.size for param in self._params.values()])
        px_gb_estimate = FUDGE_FACTOR * PARAM_SIZE_IN_GB * num_params
        # whats the largest batch size B such that B * px_gb_estimate < 16?
        largest_batch = GB_LIMIT / px_gb_estimate
        if self._hyperparams['lot_size'] <= largest_batch:
            return self._hyperparams['lot_size']
        # what if this largest_batch size leaves a very uneven final batch?
        num_batches = math.ceil(self._hyperparams['lot_size'] / largest_batch)
        return math.ceil(self._hyperparams['lot_size'] / num_batches)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import pickle
from mxnet import nd


def import_adult_dataset(data_ctx):
    try:
        Xtrain, Ytrain = pickle.load( open( "dpareto/data/adult/a1a.train.p", "rb" ) )
        Xtest, Ytest = pickle.load( open( "dpareto/data/adult/a1a.test.p", "rb" ) )
        return Xtrain, Ytrain, Xtest, Ytest
    except FileNotFoundError:
        print('Adult dataset files not found. Try running the downloader from the project root.')



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import mxnet as mx

from dpareto.models.adult.base import AdultBase


class AdultLrBase(AdultBase):
    def __init__(self, hyperparams, options={}):
        super(AdultLrBase, self).__init__(hyperparams, options)

    @staticmethod
    def _get_loss_func():
        return mx.gluon.loss.SigmoidBinaryCrossEntropyLoss()



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import mxnet as mx
from mxnet import nd
import numpy as np
# from pydiffpriv import dpacct
from autodp import rdp_acct
import time

from dpareto.utils.object_io import save_object
from dpareto.utils.progress_bar import printProgressBar


class DpFeedforwardNet:
    def __init__(self, hyperparams, options={}):
        self._hyperparams = hyperparams
        self._epochs = hyperparams['epochs']  # total number of 'virtual epochs' to run
        self._lot_size = hyperparams['lot_size']  # lot_size as in DP-SGD paper
        self._fixed_delta = hyperparams['fixed_delta']  # for (eps, delta)-DP, this is the fixed value for delta

        # misc options and their defaults (if unspecified)
        self._options = options
        self._optimizer_name = options.get('optimizer_name', 'dp_sgd')
        self._enable_mxnet_profiling = options.get('enable_mxnet_profiling', False)
        self._save_plots = options.get('save_plots', False)
        self._print_epoch_status = options.get('print_epoch_status', True)
        self._compute_epoch_accuracy = options.get('compute_epoch_accuracy', True)
        self._verbose = options.get('verbose', False)
        self._use_gpu = options.get('use_gpu', True)
        self._run_training = options.get('run_training', True)
        self._debugging = options.get('debugging', False)
        self._accumulate_privacy = options.get('accumulate_privacy', True)
        self._output_base_dir = options.get('output_base_dir', '')

        # TODO: switching between optimizers this way doesn't feel right
        if self._optimizer_name == 'dp_sgd':
            from dpareto.optimizers.dp_sgd import DpSgd as Optimizer
        elif self._optimizer_name == 'dp_adam':
            from dpareto.optimizers.dp_adam import DpAdam as Optimizer
        else:
            raise NotImplementedError("Optimizer not implemented.")
        self._optimizer = Optimizer

        # MXNet settings
        if self._use_gpu:
            ctx = mx.gpu(options['device_id']) if options.get('device_id', False) else mx.gpu()
        else:
            ctx = mx.cpu()
        self._data_ctx = ctx
        self._model_ctx = ctx

        self._name = 'dp_ffnn'

    def run(self):
        # Helper methods
        def get_random_lot(data_loader):
            return next(iter(data_loader))

        # Data importing, pre-processing, and loading
        num_training_examples, num_testing_examples, train_data_lot_iterator, train_data_eval_iterator, test_data = self._load_data()
        # parameters calculated from loaded data
        self._num_training_examples = num_training_examples
        self._num_testing_examples = num_testing_examples
        self._hyperparams['sample_fraction'] = self._lot_size / num_training_examples
        rounds_per_epoch = round(num_training_examples / self._lot_size)

        # Set up privacy accountant
        accountant = rdp_acct.anaRDPacct() # dpacct.anaCGFAcct()
        eps_sequence = []

        # Network structure creation
        self._create_network_params()

        # Loss function
        loss_func = self._get_loss_func()

        # Optimization procedure
        trainer = self._optimizer(self._hyperparams, self._net, self._params, loss_func, self._model_ctx, accountant)

        # begin profiling if enabled
        if self._enable_mxnet_profiling:
            from mxnet import profiler
            profiler.set_config(profile_all=True, aggregate_stats=True, filename='profile_output.json')
            profiler.set_state('run')

        # Training sequence
        rounds = round(self._epochs * rounds_per_epoch)
        loss_sequence = []
        current_epoch_loss = mx.nd.zeros(1, ctx=self._model_ctx)
        for t in range(1, rounds + 1):
            if self._verbose and self._print_epoch_status:
                # show current epoch progress
                epoch_number = 1 + (t - 1) // rounds_per_epoch
                epoch_progress = 1 + (t - 1) % rounds_per_epoch
                printProgressBar(epoch_progress, rounds_per_epoch, prefix='Epoch {} progress:'.format(epoch_number), length=50)

            if self._run_training:
                # prepare random lot of data for DPSGD step
                data, labels = get_random_lot(train_data_lot_iterator)
                data = data.as_in_context(self._model_ctx).reshape((-1, 1, self._input_layer))
                labels = labels.as_in_context(self._model_ctx)
            else:
                data, labels = [], []

            # perform DPSGD step
            lot_mean_loss = trainer.step(data, labels, accumulate_privacy=self._accumulate_privacy, run_training=self._run_training)

            loss_sequence.append(lot_mean_loss)
            current_epoch_loss += lot_mean_loss

            # no need to continue running training if NaNs are present
            if not np.isfinite(lot_mean_loss):
                self._run_training = False
                if self._verbose: print("NaN loss on round {}.".format(t))
            if self._params_not_finite():
                self._run_training = False
                if self._verbose: print("Non-finite parameters on round {}.".format(t))


            if self._accumulate_privacy and self._debugging:
                eps_sequence.append(accountant.get_eps(self._fixed_delta))

            # print some stats after an "epoch"
            if t % rounds_per_epoch == 0:
                if self._verbose:
                    print("Epoch {}  (round {})  complete.".format(t / rounds_per_epoch, t))
                    if self._run_training:
                        print("mean epoch loss: {}".format(current_epoch_loss.asscalar() * self._lot_size / self._num_training_examples))
                        if self._compute_epoch_accuracy:
                            print("training accuracy: {}".format(self._evaluate_accuracy(train_data_eval_iterator)))
                            print("testing accuracy: {}".format(self._evaluate_accuracy(test_data)))
                    if self._accumulate_privacy and self._debugging:
                        print("eps used: {}\n".format(eps_sequence[-1]))
                    print()
                current_epoch_loss = mx.nd.zeros(1, ctx=self._model_ctx)

        # end profiling if enabled
        if self._enable_mxnet_profiling:
            mx.nd.waitall()
            profiler.set_state('stop')
            print(profiler.dumps())

        # Make sure we don't report a bogus number
        if self._accumulate_privacy:
            final_eps = accountant.get_eps(self._fixed_delta)
        else:
            final_eps = -1

        test_accuracy = self._evaluate_accuracy(test_data)

        if self._save_plots or self._debugging:
            self._create_and_save_plots(t, eps_sequence, loss_sequence, final_eps, test_accuracy)

        return final_eps, test_accuracy

    def _create_network_params(self):
        # Set the base parameters for the network
        self._params = {}
        if len(self._hidden_layers) > 0:
            #  Allocate parameters for input layer -> first hidden layer
            self._params['W0'] = self._xavier_initialize_params(self._input_layer, self._hidden_layers[0], self._model_ctx)
            self._params['b0'] = self._xavier_initialize_params(1, self._hidden_layers[0], self._model_ctx)

            #  Allocate parameters for hidden layer i -> hidden layer i+1
            for i in range(1, len(self._hidden_layers)):
                str_index = str(i)
                self._params['W' + str_index] = self._xavier_initialize_params(self._hidden_layers[i-1], self._hidden_layers[i], self._model_ctx)
                self._params['b' + str_index] = self._xavier_initialize_params(1, self._hidden_layers[i], self._model_ctx)

            #  Allocate parameters for final hidden layer -> output layer
            str_index = str(len(self._hidden_layers))
            self._params['W' + str_index] = self._xavier_initialize_params(self._hidden_layers[-1], self._output_layer, self._model_ctx)
            self._params['b' + str_index] = self._xavier_initialize_params(1, self._output_layer, self._model_ctx)
        else:
            #  Allocate parameters for input layer -> output layer
            self._params['W0'] = self._xavier_initialize_params(self._input_layer, self._output_layer, self._model_ctx)
            self._params['b0'] = self._xavier_initialize_params(1, self._output_layer, self._model_ctx)

    @staticmethod
    def _xavier_initialize_params(num_in, num_out, context, magnitude=3):
        params_shape = (num_in, num_out)
        weight_scale = np.sqrt(magnitude * 2 / (num_in + num_out))
        return nd.random_normal(shape=params_shape, scale=weight_scale, ctx=context)

    # Main network function to compute forward pass of the data.
    def _net(self, in_data, params, training_mode=True):
        if training_mode:
            dot = nd.batch_dot
        else:
            dot = nd.dot

        # if there are no hidden layers, just compute the output as a linear combination
        if len(self._hidden_layers) == 0:
            return (dot(in_data, params['W0']) + params['b0']).reshape((in_data.shape[0], self._output_layer))

        #  Compute the first hidden layer
        h0_linear = dot(in_data, params['W0']) + params['b0']
        h0 = nd.relu(h0_linear)

        #  Compute the (i+1)^th hidden layer
        hprevious = h0
        for i in range(1, len(self._hidden_layers)):
            str_index = str(i)
            hcurrent_linear = dot(hprevious, params['W' + str_index]) + params['b' + str_index]
            hcurrent = nd.relu(hcurrent_linear)
            hprevious = hcurrent

        str_index = str(len(self._hidden_layers))
        yhat_linear = dot(hprevious, params['W' + str_index]) + params['b' + str_index]

        return yhat_linear.reshape((in_data.shape[0], self._output_layer))

    def _params_not_finite(self):
        for param in self._params.values():
            if not np.isfinite(param.asnumpy()).all():
                return True
        return False

    @staticmethod
    def _get_loss_func():
        raise NotImplementedError("Method not implemented.")

    def _load_data(self):
        raise NotImplementedError("Method not implemented.")

    def _evaluate_accuracy(self, data_iterator):
        raise NotImplementedError("Method not implemented.")

    @staticmethod
    def moving_average(a, n=3):
        ret = np.cumsum(a, dtype=float)
        ret[n:] = ret[n:] - ret[:-n]
        return ret[n - 1:] / n

    # Plot and save results
    def _create_and_save_plots(self, rounds, eps_sequence, loss_sequence, final_eps, test_accuracy):
        # Estabilish output directory
        dir_name = np.abs(hash((time.time(), frozenset(self._hyperparams.items()))))
        if not self._output_base_dir:
            plots_subdir = self._name
            plots_dir = 'results/instance-results/{}/{}'.format(plots_subdir, dir_name)
        else:
            plots_dir = '{}/instance-results/{}'.format(self._output_base_dir, dir_name)
        import os
        os.makedirs(plots_dir, exist_ok=True)

        # Save hyperparameter and results info with plots
        final_results = (self._hyperparams, final_eps, test_accuracy)
        save_object(plots_dir, final_results, 'final_results')

        import matplotlib.pyplot as plt  # import here for weird matplotlib+multiprocessign compatibility issue

        # Only create privacy plot if privacy was accumulated
        if len(eps_sequence) > 0:
            # Privacy results
            plt.figure(num=1, figsize=(12, 8), dpi=80, facecolor='w', edgecolor='k')
            plt.grid(True, which='both')
            plt.plot(range(rounds), eps_sequence)
            plt.xlabel('rounds', fontsize=14)
            plt.ylabel('eps', fontsize=14)
            plt.title('Overall (eps,delta)-DP over composition.')
            plt.savefig('{}/privacy-results.png'.format(plots_dir))
            plt.close()

        # Loss results
        plt.figure(num=None, figsize=(8, 6))
        plt.plot(range(len(loss_sequence)), loss_sequence)
        ma_window = 50
        plt.plot(range(ma_window-1, len(loss_sequence)), self.moving_average(loss_sequence, n=ma_window), color='red')
        plt.grid(True, which='both')
        plt.xlabel('\"rounds\"', fontsize=14)
        plt.ylabel('average loss', fontsize=14)
        plt.draw()
        plt.savefig('{}/loss-results.png'.format(plots_dir))
        plt.close()

        # # Gradient clipping results
        # plt.figure(num=None,figsize=(8, 6))
        # plt.plot(norm_sequence)
        # plt.grid(True, which="both")
        # plt.xlabel('rounds',fontsize=14)
        # plt.ylabel('gradient norm',fontsize=14)
        # plt.draw()
        # plt.savefig("{}/dp-lr-norm-results.png".format(plots_dir))



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import mxnet as mx
from mxnet import nd

from dpareto.optimizers.dp_optimizer import DpOptimizer

class DpSgd(DpOptimizer):
    def __init__(self, hyperparams, net, params, loss_func, model_ctx, accountant):
        super(DpSgd, self).__init__(hyperparams, net, params, loss_func, model_ctx, accountant)

        # Compute scale of Gaussian noise to add
        self._hyperparams['sigma'] = hyperparams['z'] * (2 * hyperparams['l2_clipping_bound'] / hyperparams['lot_size'])

    def _update_params(self, accumulated_grads):
        # scale gradients by lot size, add noise, and update the parameters
        for param_name, param in self._params.items():
            # average the clipped gradients and then add noise to each averaged gradient
            param_grad_update = (accumulated_grads[param_name] / self._hyperparams['lot_size']) + \
                                mx.random.normal(0, self._hyperparams['sigma'], param.shape, ctx=self._model_ctx)

            # update params with SGD
            param[:] = nd.sgd_update(weight=param, grad=param_grad_update, lr=self._hyperparams['lr'])



import mxnet as mx
from mxnet import nd, gluon
import numpy as np

import dpareto.data.adult.importer as importer
from dpareto.models.adult.base import AdultBase
from dpareto.utils.lot_sampler import LotSampler


class DpAdultSvmSgd(AdultBase):
    def __init__(self, hyperparams, options={}):
        super(DpAdultSvmSgd, self).__init__(hyperparams, options)

        self._input_layer = 123
        self._hidden_layers = []
        self._output_layer = 1

        self._name = 'dp_adult_svm_sgd'

    def _load_data(self):
        xTrain, yTrain, xTest, yTest = importer.import_adult_dataset(self._data_ctx)
        yTrain = yTrain * 2 - 1
        yTest = yTest * 2 - 1

        num_training_examples = len(yTrain)
        num_testing_examples = len(yTest)

        if self._verbose:
            print("Loading...")
        train_data = gluon.data.ArrayDataset(xTrain, yTrain)
        train_data_lot_iterator = gluon.data.DataLoader(train_data,
                                                        batch_sampler=LotSampler(self._lot_size, num_training_examples))
        train_data_eval_iterator = gluon.data.DataLoader(train_data, batch_size=self._lot_size)
        test_data = gluon.data.DataLoader(gluon.data.ArrayDataset(xTest, yTest),
                                          batch_size=self._lot_size, shuffle=True)
        if self._verbose:
            print("done loading.")

        return num_training_examples, num_testing_examples, train_data_lot_iterator, train_data_eval_iterator, test_data

    @staticmethod
    def _get_loss_func():
        return mx.gluon.loss.HingeLoss()

    def _evaluate_accuracy(self, data_iterator):
        num_correct = 0
        total = 0
        for i, (data, label) in enumerate(data_iterator):
            data = data.as_in_context(self._model_ctx).reshape((-1, 1, self._input_layer))
            label = label.as_in_context(self._model_ctx)

            output = self._net(data, self._params, training_mode=False)
            prediction = (output > 0.0) * 2 - 1

            num_correct += nd.sum(prediction == label)
            total += len(label)

        return num_correct.asscalar() / total


# Small run for testing purposes.
def main():
    print("Running dp_adult_svm_sgd's main().")

    # Fixing some combo of these random seeds is useful for debugging. I'll fix them all to be safe.
    import random
    random_seed = 112358
    random.seed(random_seed)
    np.random.seed(random_seed)
    mx.random.seed(random_seed)

    # Some default hyperparameter setting.
    hyperparams = {'epochs': 1,
                   'lot_size': 64,
                   'lr': 0.05,
                   'l2_clipping_bound': 2.0,
                   'z': 1.0,
                   'fixed_delta': 1e-5}
    print("hyperparams: {}".format(hyperparams))

    options = {'use_gpu': False, 'verbose': True, 'accumulate_privacy': True, 'debugging': True}

    instance = DpAdultSvmSgd(hyperparams, options)
    priv, acc = instance.run()
    print("Instance privacy: {}".format((priv, hyperparams['fixed_delta'])))
    print("Instance accuracy: {}".format(acc))


if __name__ == "__main__":
    main()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()


setuptools.setup(
  name = 'dpareto',
  version = '0.1.2',
  license='Apache 2.0',
  description = 'Automatic Discovery of Privacy-Utility Pareto Fronts',
  long_description=long_description,
  author = 'Amazon',
  author_email = 'tdiethe@amazon.com',
  url = 'https://github.com/amzn/differential-privacy-bayesian-optimization',
  download_url = 'https://github.com/amzn/differential-privacy-bayesian-optimization/archive/v_01.tar.gz',
  keywords = ['Differential privacy', 'Bayesian optimization'],
  packages = setuptools.find_packages(),
  install_requires=[
          'gpflow @ git+https://github.com/GPflow/GPflow.git@ce5ad7ea75687fb0bf178b25f62855fc861eb10f',
          'gpflowopt @ git+https://github.com/GPflow/GPflowOpt.git',
          'seaborn',
          'psutil',
          'mxnet',
          'autodp',
      ],
  classifiers=[
    'Development Status :: 3 - Alpha',
    'Intended Audience :: Science/Research',
    'Topic :: Scientific/Engineering',
    'License :: OSI Approved :: Apache Software License',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
  ],
)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import mxnet as mx


def get_gpu_count():
    max_gpus = 64
    for i in range(max_gpus):
        try:
            mx.nd.zeros((1,), ctx=mx.gpu(i))
        except:
            return i
    return max_gpus


# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import random
import mxnet as mx
import numpy as np


def set_random_seed(value):
    wrapped_value =  value % (2**32 - 1)
    random.seed(wrapped_value)
    np.random.seed(wrapped_value)
    mx.random.seed(wrapped_value)